# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
import dataiku
import pandas as pd
import numpy as np
from datetime import datetime
from pulp import LpVariable, LpProblem, LpMinimize, LpStatus, lpSum, value
import re

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Import Trucks
trucks_dataset = dataiku.Dataset("trucks_slots_joined_1_prepared")
trucks = trucks_dataset.get_dataframe()
trucks_and_docks_and_slots_list = trucks.truck_and_dock_and_slot.tolist()
trucks.set_index('truck_and_dock_and_slot', inplace=True)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Import slot prices for the end of the notebook
compatibility_dataset = dataiku.Dataset("input_compatibility")
compatibility = compatibility_dataset.get_dataframe()
compatibility["price"] = compatibility["slot_weight"]

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Import slot availability for the end of the notebook
slot_availability_dataset = dataiku.Dataset("slots_availability_joined_windows_prepared")
slot_availability = slot_availability_dataset.get_dataframe()

# Check if some slots have been selected
if len(slot_availability)==0:
    raise Exception("Please select some opened slots in the webapp!")

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Define Decision Variables
docks_and_slots_list = trucks.docks_and_slots.tolist()
slots_list = trucks.slot.unique().tolist()
docks_list = trucks.dock.unique().tolist()
docks_list = list(map(str, docks_list))
trucks_list = trucks.truck_name_arrival_ts_first.unique().tolist()#because one truck can arrive several times during 1 day at the warehouse

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Initialize Class
model = LpProblem("Slots_Optimization", LpMinimize)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Create Decision Variables
# Assign a truck to a slot
x = LpVariable.dicts("allocation", [tuple(k.split(',')) for k in trucks_and_docks_and_slots_list], cat='Binary')#version 04/08

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Define Objective Function
model += lpSum(trucks.loc[k, "waiting_cost_predicted"]*x[tuple(k.split(','))] for k in trucks_and_docks_and_slots_list)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Build Keys that exist in the Variable x
docks_and_slots_dict = dict()
for i in trucks_list:
    docks_and_slots_dict[i] = list()
    for dock in docks_list:
        for slot in slots_list:

            if (i, dock, slot) in x:
                docks_and_slots_dict[i].append((dock, slot))

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Build Keys that exist to get trucks possible for 1 (dock,slot)
trucks_for_docks_slots_dict = dict()

for dock in docks_list:
    for slot in slots_list:
        trucks_for_docks_slots_dict[(dock, slot)] = list()
        for truck in trucks_list:
            if (truck, dock, slot) in x:
                trucks_for_docks_slots_dict[(dock, slot)].append(truck)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Add Constraints
for i in trucks_list:
    model += lpSum(x[(i, j[0], j[1])] for j in docks_and_slots_dict[i]) >= 1
    model += lpSum(x[(i, j[0], j[1])] for j in docks_and_slots_dict[i]) <= 1
for dock_slot, trucks in trucks_for_docks_slots_dict.items():
    model += lpSum(x[(i, dock_slot[0], dock_slot[1])] for i in trucks) <= 1

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Solve Model
model.solve()
print("Total Costs = {:,}".format(int(value(model.objective))))
print('\n' + "Status: {}".format(LpStatus[model.status]))

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Results of the optimization: Allocation of trucks to slots and docks
storage_dict = {"truck":[],"dock":[],"slot":[],"dock_slot":[],"assigned":[]};

for v in model.variables():
    if v.varValue > 0:
        buffer = re.findall("([^']*)", v.name)
        storage_dict["truck"].append(buffer[2])
        storage_dict["dock"].append(buffer[6])
        storage_dict["slot"].append(buffer[10])
        storage_dict["dock_slot"].append([buffer[6],buffer[10]])
        storage_dict["assigned"].append(v.varValue)

# Create the allocation dataset
storage_df = pd.DataFrame(storage_dict, columns=['truck','dock','slot','dock_slot', 'assigned'])

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Calculate allocated slots price and the corresponding warehouse inefficiencies
allocated_docks_slots = list()
cost_open_dock_slot = 0
total_availability_cost = 0

for dispo in storage_df.dock_slot:
    cost_open_dock_slot += compatibility['price'][compatibility['dock'] == int(dispo[0][0])].values[0]

for i in slot_availability.dock.values:
    total_availability_cost += compatibility['price'][compatibility['dock'] == i].values[0]

inefficiency = total_availability_cost - cost_open_dock_slot

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Get current time
time = datetime.now()

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
# Create the total_costs dataframe

# Get the user who launched the scenario:

project_key = dataiku.default_project_key()
project = dataiku.api_client().get_project(project_key) 
scenario = project.get_scenario("FLOWREBUILD+OPTIMIZATION")
current_run = scenario.get_current_run()

if current_run == None: 
    client = dataiku.api_client()
    auth_info = client.get_auth_info()
    username= auth_info["authIdentifier"].split('@')[0]
else:    
    current_run_log = current_run.get_info()
    run_full_log = scenario.get_run(current_run_log["runId"])
    info = run_full_log.get_info()
    username = info['trigger']['triggerAuthCtx']['realUserLogin'].split('@')[0]


    
total_costs = {'Scenario ID': np.random.randint(low=1e5, high=1e6),
               'User': username,
               'Date Time': datetime.now(),
               'Total costs': [int(value(model.objective))],
               'Total costs of non used docks which are available': [int(inefficiency)],
               'Solver Status': LpStatus[model.status]
              }
total_costs_df = pd.DataFrame(data=total_costs)

# -------------------------------------------------------------------------------- NOTEBOOK-CELL: CODE
#clean output
colums_to_keep=['truck', 'dock', 'slot']
storage_df=storage_df[colums_to_keep]
storage_df=storage_df.dropna()


# Recipe outputs
total_costs = dataiku.Dataset("total_costs")
total_costs.write_with_schema(total_costs_df)

storage_optim = dataiku.Dataset("dock_optim")
storage_optim.write_with_schema(storage_df)